import { randomNumber, spawnEntityInRegion, BlockManager, VFX, calculateDistance, equipRandomItemsToZombie, isAllowedBlock, allowedSuffixes_list, allowedTags_list } from "./functions.js";
import { EntityComponentTypes, EquipmentSlot, ItemStack, world } from "@minecraft/server";

export class HerobrineTriggers {
    constructor(targetSelected, herobrineData, herobrineSettings, Dimension) {
        this.targetSelected = targetSelected;
        this.herobrineData = herobrineData;
        this.herobrineSettings = herobrineSettings;
        this.Dimension = Dimension;
    }

    darkenVision() {
        this.targetSelected.camera.fade({
            fadeTime: { fadeInTime: 1, holdTime: 3, fadeOutTime: 1 },
            fadeColor: { blue: 0, green: 0, red: 0 }
        });

        this.targetSelected.onScreenDisplay.setTitle(
            { text: " " },
            {
                fadeInDuration: 20,
                fadeOutDuration: 20,
                stayDuration: 50,
                subtitle: {
                    rawtext: [
                        { text: "§c" },
                        { translate: `herobrine.message.scared_${randomNumber(1, 16)}` }
                    ]
                }
            }
        );
    }

    herobrineCrash() {
        this.herobrineCrash();
    }

    ownAnimals() {
        const entityOwnList = {
            "minecraft:pig": "hrb:pig",
            "minecraft:cow": "hrb:cow",
            "minecraft:sheep": "hrb:sheep",
            "minecraft:chicken": "hrb:chicken"
        };

        const entitiesInPlayerLoc = this.Dimension.getEntities({
            location: this.targetSelected.location,
            maxDistance: randomNumber(30, 45)
        });

        entitiesInPlayerLoc.forEach(entityInRegion => {
            const transformEntityID = entityOwnList[entityInRegion.typeId];
            if (transformEntityID) {
                const ownSpawn = this.Dimension.spawnEntity(transformEntityID, entityInRegion.location);
                ownSpawn.setRotation(entityInRegion.getRotation());
                entityInRegion.remove();
            }
        });
    }

    torchTransform() {
        const blockManager = new BlockManager();

        if (this.targetSelected.location.y > 45 && randomNumber(1, 4) === 1) {
            blockManager.squareGetBlocks({
                locationStart: this.targetSelected.location,
                allSides: false,
                blockTargetID: "minecraft:torch"
            });
            return;
        }

        if (this.targetSelected.location.y < 45 && randomNumber(1, 2) == 1) {
            blockManager.squareGetBlocks({
                locationStart: this.targetSelected.location,
                allSides: false,
                blockTargetID: "minecraft:torch"
            });
            if (randomNumber(1, 2) == 1) this.scarySound()
        }
    }

    tunnelGeneration() {
        if (randomNumber(1, 5) == 1) {
            // Limite mínimo para a coordenada Y
            const minYCoordinate = -60;

            // Verifica se os dois blocos (inferior e superior) na posição são "minecraft:air"
            const areBothBlocksAir = (position) => {
                const lowerBlock = this.Dimension.getBlock(position);
                const upperBlock = this.Dimension.getBlock({
                    x: position.x,
                    y: position.y + 1,
                    z: position.z
                });
                return (lowerBlock?.typeId === "minecraft:air" && !lowerBlock?.isLiquid) && (upperBlock?.typeId === "minecraft:air" && !upperBlock?.isLiquid);
            };

            // Percorre a trajetória e retorna o primeiro ponto onde pelo menos um dos blocos não é "minecraft:air"
            const findObstaclePosition = (origin, direction, range = 60) => {
                for (let i = 0; i < range; i++) {
                    const pos = {
                        x: origin.x + direction.x * i,
                        y: origin.y,
                        z: origin.z + direction.z * i
                    };
                    if (!areBothBlocksAir(pos)) {
                        return pos;
                    }
                }
                return null;
            };

            // Gera a trajetória do túnel, incluindo os blocos do chão e a camada superior
            const generateTunnelPath = (origin, direction, range = 24) => {
                const path = [];
                for (let i = 0; i < range; i++) {
                    const basePos = {
                        x: origin.x + direction.x * i,
                        y: origin.y,
                        z: origin.z + direction.z * i
                    };
                    // Adiciona a camada inferior e a superior para um túnel com 2 blocos de altura
                    path.push(basePos);
                    path.push({
                        x: basePos.x,
                        y: basePos.y + 1,
                        z: basePos.z
                    });
                }
                return path;
            };

            // Verifica se há um bloco sólido na direção indicada, que servirá de suporte para a tocha
            const hasSolidBlockInDirection = (pos, faceDirection) => {
                const offset = { x: 0, y: 0, z: 0 };
                switch (faceDirection) {
                    case "north": offset.z = -1; break;
                    case "south": offset.z = 1; break;
                    case "east": offset.x = 1; break;
                    case "west": offset.x = -1; break;
                }
                const adjacentPos = {
                    x: pos.x + offset.x,
                    y: pos.y,
                    z: pos.z + offset.z
                };
                const adjacentBlock = this.Dimension.getBlock(adjacentPos);
                // Retorna true se o bloco existir, não for "minecraft:air", não for líquido e não for um bloco permitido
                return !!adjacentBlock &&
                    adjacentBlock.typeId !== "minecraft:air" &&
                    !adjacentBlock.isLiquid &&
                    !isAllowedBlock(adjacentBlock, allowedSuffixes_list, allowedTags_list);
            };

            // Lista de direções possíveis para o túnel
            const directions = [
                { x: 1, y: 0, z: 0 },
                { x: -1, y: 0, z: 0 },
                { x: 0, y: 0, z: 1 },
                { x: 0, y: 0, z: -1 }
            ];
            const selectedDirection = directions[randomNumber(0, directions.length - 1)];

            // Define o ponto de início do túnel com limite mínimo na coordenada Y
            let startY = this.targetSelected.location.y;
            startY = Math.max(startY, minYCoordinate); // Garante que startY não seja menor que minYCoordinate

            const obstaclePos = findObstaclePosition({ x: this.targetSelected.location.x, y: startY, z: this.targetSelected.location.z }, selectedDirection, 50);
            const startPos = obstaclePos ? obstaclePos : { x: this.targetSelected.location.x, y: startY, z: this.targetSelected.location.z };


            // Gera e "escava" o túnel definindo cada posição do caminho como "minecraft:air"
            const tunnelPath = generateTunnelPath(startPos, selectedDirection, 50);
            tunnelPath.forEach(pos => {
                this.Dimension.setBlockType(pos, "minecraft:air");
            });

            // Gera a trajetória para posicionar as tochas (na camada superior)
            const torchPositions = [];
            for (let i = 0; i < 50; i++) {
                torchPositions.push({
                    x: startPos.x + selectedDirection.x * i,
                    y: startPos.y + 1,
                    z: startPos.z + selectedDirection.z * i
                });
            }

            // A cada 4 blocos, verifica se há suporte em alguma lateral para fixar a tocha
            torchPositions.forEach((pos, index) => {
                if (index % 4 === 0) {
                    let torchDirection = null;
                    // Se o túnel for orientado no eixo X, procura suporte nas laterais norte e sul
                    if (selectedDirection.x !== 0) {
                        if (hasSolidBlockInDirection(pos, "north")) {
                            torchDirection = "north";
                        } else if (hasSolidBlockInDirection(pos, "south")) {
                            torchDirection = "south";
                        }
                    }
                    // Se for orientado no eixo Z, procura suporte nas laterais leste e oeste
                    else if (selectedDirection.z !== 0) {
                        if (hasSolidBlockInDirection(pos, "east")) {
                            torchDirection = "east";
                        } else if (hasSolidBlockInDirection(pos, "west")) {
                            torchDirection = "west";
                        }
                    }
                    if (torchDirection) {
                        this.Dimension.setBlockType(pos, "minecraft:redstone_torch");
                        const torchBlock = this.Dimension.getBlock(pos);
                        if (torchBlock) {
                            torchBlock.setPermutation(
                                torchBlock.permutation.withState("torch_facing_direction", torchDirection)
                            );
                        }
                    }
                }
            });
        }
    }

    scarySound() {
        const soundsInMiner = [
            "scary.dirt_breaking.sound",
            "scary.stone_breaking.sound",
        ]
        const soundsDefault = [
            "ambient.basalt_deltas.mood",
            "ambient.cave",
            "ambient.crimson_forest.mood",
            "ambient.warped_forest.mood"
        ];

        if (this.targetSelected.location.y < 40) {
            if (randomNumber(1, 20) == 1) {
                this.targetSelected.playSound(soundsInMiner[1]);
            } else {
                this.targetSelected.playSound(soundsDefault[randomNumber(0, soundsDefault.length - 1)]);
            }
        } else if (this.targetSelected.location.y > 40 && this.targetSelected.location.y < 90) {
            if (randomNumber(1, 10) == 1) {
                this.targetSelected.playSound(soundsInMiner[0]);
            } else {
                this.targetSelected.playSound(soundsDefault[randomNumber(0, soundsDefault.length - 1)]);
            }
        } else {
            this.targetSelected.playSound(soundsDefault[randomNumber(0, soundsDefault.length - 1)]);
        }
    }

    signSpawn(blockID = "minecraft:standing_sign", loc = this.targetSelected.location, rad = 8, height = 2, minDistance = 4) {
        try {

            const generateCoordinatesWithinVolume = () => {
                const validPositions = [];

                for (let dx = -rad; dx <= rad; dx++) {
                    for (let dz = -rad; dz <= rad; dz++) {
                        for (let dy = -height; dy <= height; dy++) {
                            const spX = Math.round(loc.x + dx);
                            const spY = Math.round(loc.y + dy);
                            const spZ = Math.round(loc.z + dz);

                            const isFarEnough =
                                Math.abs(spX - loc.x) >= minDistance &&
                                Math.abs(spZ - loc.z) >= minDistance;

                            if (!isFarEnough) continue;

                            const blocks = [
                                this.Dimension.getBlock({ x: spX, y: spY, z: spZ }),
                                this.Dimension.getBlock({ x: spX, y: spY - 1, z: spZ })
                            ];

                            const [currentBlock, belowBlock] = blocks;

                            if (isAllowedBlock(belowBlock)) continue;

                            const isValidLocation =
                                currentBlock.typeId === "minecraft:air" &&
                                belowBlock.typeId !== "minecraft:air" &&
                                !currentBlock.isLiquid &&
                                !belowBlock.isLiquid;

                            if (isValidLocation) {
                                validPositions.push({ x: spX, y: spY, z: spZ });
                            }
                        }
                    }
                }
                return validPositions;
            };

            const validPositions = generateCoordinatesWithinVolume();

            if (validPositions.length > 0) {
                const randomIndex = Math.floor(Math.random() * validPositions.length);
                const selectedPosition = validPositions[randomIndex];

                this.Dimension.setBlockType({
                    x: selectedPosition.x,
                    y: selectedPosition.y,
                    z: selectedPosition.z
                }, blockID);

                const blockInLocation = this.Dimension.getBlock(selectedPosition);
                const signComponent = blockInLocation.getComponent("minecraft:sign");

                this.Dimension.setBlockPermutation(
                    blockInLocation,
                    blockInLocation.permutation.withState("ground_sign_direction", randomNumber(0, 15))
                );

                signComponent.setText({
                    rawtext: [
                        { text: "§l§c" },
                        { translate: `herobrine.sign.message_${randomNumber(1, 15)}` }
                    ]
                });

                signComponent.setWaxed(true);
                return;
            }

            console.warn("Não foi possível encontrar uma posição válida para colocar a placa.");
            return null;

        } catch (error) {
            console.error("Erro ao tentar colocar a placa:", error);
            return null;
        }
    }

    herobrineChat() {
        world.sendMessage({
            rawtext: [
                { text: `<` }, { translate: "entity.hrb:herobrine.name" }, { text: "> " },
                { translate: `herobrine.herobrine_chat.message_${randomNumber(1, 10)}` }
            ]
        });
    }

    randomEvent() {
        const locationFix = {
            x: Math.round(this.herobrineData.location.x),
            y: Math.round(this.herobrineData.location.y),
            z: Math.round(this.herobrineData.location.z)
        };

        switch (randomNumber(1, 4)) {
            case 1:
                if (this.herobrineSettings.world_vandalism) {
                    this.Dimension.spawnEntity("minecraft:lightning_bolt", locationFix);
                }
                break;

            case 2:
                if (this.herobrineSettings.world_vandalism) {
                    for (let i = 0; i < randomNumber(1, 5); i++) {
                        this.Dimension.spawnEntity("minecraft:silverfish", locationFix);
                    }
                } else {
                    this.scarySound();
                }
                break;

            case 3:
                for (let i = 0; i < randomNumber(1, 5); i++) {
                    this.Dimension.spawnEntity("minecraft:zombie", locationFix);
                }
                break;

            case 4:
                if (this.herobrineSettings.world_vandalism) {
                    this.Dimension.createExplosion(locationFix, 3, { causesFire: true });
                } else {
                    this.scarySound();
                }
                break;
        }
    }
}